@@ -1,6 +1,8 @@
+/* -*- c-basic-offset: 2 -*- */
 /* 
   Copyright(C) 2010 Tetsuro IKEDA
   Copyright(C) 2010 Kentoku SHIBA
+  Copyright(C) 2011 Kouhei Sutou <kou@clear-code.com>
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
@@ -815,7 +817,8 @@ void mrn_ft_close_search(FT_INFO *handler)
 
 /* handler implementation */
 ha_mroonga::ha_mroonga(handlerton *hton, TABLE_SHARE *share)
-  :handler(hton, share)
+  :handler(hton, share),
+   ignoring_duplicated_key(false)
 {
   DBUG_ENTER("ha_mroonga::ha_mroonga");
   ctx = grn_ctx_open(0);
@@ -872,7 +875,6 @@ ulonglong ha_mroonga_table_flags =
     HA_STATS_RECORDS_IS_EXACT |
     HA_NO_PREFIX_CHAR_KEYS |
     HA_CAN_FULLTEXT |
-    HA_NO_AUTO_INCREMENT |
     HA_CAN_INSERT_DELAYED |
     HA_BINLOG_FLAGS |
     HA_CAN_BIT_FIELD |
@@ -1422,6 +1424,22 @@ void ha_mroonga::position(const uchar *record)
   DBUG_VOID_RETURN;
 }
 
+int ha_mroonga::extra(enum ha_extra_function operation)
+{
+  DBUG_ENTER("ha_mroonga::extra");
+  switch (operation) {
+  case HA_EXTRA_IGNORE_DUP_KEY:
+    ignoring_duplicated_key = true;
+    break;
+  case HA_EXTRA_NO_IGNORE_DUP_KEY:
+    ignoring_duplicated_key = false;
+    break;
+  default:
+    break;
+  }
+  DBUG_RETURN(0);
+}
+
 int ha_mroonga::write_row(uchar *buf)
 {
   DBUG_ENTER("ha_mroonga::write_row");
@@ -1432,6 +1450,14 @@ int ha_mroonga::write_row(uchar *buf)
   THD *thd = ha_thd();
   int i, col_size;
   int n_columns = table->s->fields;
+  int error = 0;
+
+  if (table->next_number_field && buf == table->record[0])
+  {
+    if ((error = update_auto_increment()))
+      DBUG_RETURN(error);
+  }
+
 #ifndef DBUG_OFF
   my_bitmap_map *tmp_map = dbug_tmp_use_all_columns(table, table->read_set);
 #endif
@@ -1473,6 +1499,9 @@ int ha_mroonga::write_row(uchar *buf)
   int added;
   row_id = grn_table_add(ctx, tbl, pkey, pkey_size, &added);
   if (ctx->rc) {
+#ifndef DBUG_OFF
+    dbug_tmp_restore_column_map(table->read_set, tmp_map);
+#endif
     my_message(ER_ERROR_ON_WRITE, ctx->errbuf, MYF(0));
     DBUG_RETURN(ER_ERROR_ON_WRITE);
   }
@@ -1482,10 +1511,13 @@ int ha_mroonga::write_row(uchar *buf)
 #ifndef DBUG_OFF
     dbug_tmp_restore_column_map(table->read_set, tmp_map);
 #endif
+    error = HA_ERR_FOUND_DUPP_KEY;
     memcpy(dup_ref, &row_id, sizeof(grn_id));
     dup_key = pkeynr;
-    GRN_LOG(ctx, GRN_LOG_ERROR, "duplicated _id on insert");
-    DBUG_RETURN(HA_ERR_FOUND_DUPP_KEY);
+    if (!ignoring_duplicated_key) {
+      GRN_LOG(ctx, GRN_LOG_ERROR, "duplicated _id on insert");
+    }
+    DBUG_RETURN(error);
   }
 
   grn_obj colbuf;
@@ -1533,13 +1565,16 @@ int ha_mroonga::write_row(uchar *buf)
     pthread_mutex_lock(&mrn_allocated_thds_mutex);
     if (my_hash_insert(&mrn_allocated_thds, (uchar*) thd))
     {
+#ifndef DBUG_OFF
+      dbug_tmp_restore_column_map(table->read_set, tmp_map);
+#endif
       DBUG_RETURN(HA_ERR_OUT_OF_MEM);
     }
     pthread_mutex_unlock(&mrn_allocated_thds_mutex);
   }
   slot_data->last_insert_rid = row_id;
 
-  DBUG_RETURN(0);
+  DBUG_RETURN(error);
 }
 
 int ha_mroonga::update_row(const uchar *old_data, uchar *new_data)
@@ -2392,7 +2427,7 @@ void ha_mroonga::check_fast_order_limit()
 
   if (
     thd_sql_command(ha_thd()) == SQLCOM_SELECT &&
-    !select_lex->n_sum_items &&
+    !select_lex->with_sum_func &&
     !select_lex->group_list.elements &&
     !select_lex->having &&
     select_lex->table_list.elements == 1 &&
@@ -2557,9 +2592,159 @@ int ha_mroonga::reset()
     _score = NULL;
     res = NULL;
   }
+  ignoring_duplicated_key = false;
   DBUG_RETURN(0);
 }
 
+handler *ha_mroonga::clone(MEM_ROOT *mem_root)
+{
+  DBUG_ENTER("ha_mroonga::clone");
+  DBUG_RETURN(handler::clone(mem_root));
+}
+
+uint8 ha_mroonga::table_cache_type()
+{
+  DBUG_ENTER("ha_mroonga::table_cache_type");
+  DBUG_RETURN(handler::table_cache_type());
+}
+
+int ha_mroonga::read_multi_range_first(KEY_MULTI_RANGE **found_range_p,
+                                       KEY_MULTI_RANGE *ranges,
+                                       uint range_count,
+                                       bool sorted,
+                                       HANDLER_BUFFER *buffer)
+{
+  DBUG_ENTER("ha_mroonga::read_multi_range_first");
+  DBUG_RETURN(handler::read_multi_range_first(found_range_p, ranges,
+                                              range_count, sorted, buffer));
+}
+
+int ha_mroonga::read_multi_range_next(KEY_MULTI_RANGE **found_range_p)
+{
+  DBUG_ENTER("ha_mroonga::read_multi_range_next");
+  DBUG_RETURN(handler::read_multi_range_next(found_range_p));
+}
+
+void ha_mroonga::start_bulk_insert(ha_rows rows)
+{
+  DBUG_ENTER("ha_mroonga::start_bulk_insert");
+  DBUG_VOID_RETURN;
+}
+
+int ha_mroonga::end_bulk_insert()
+{
+  DBUG_ENTER("ha_mroonga::end_bulk_insert");
+  DBUG_RETURN(0);
+}
+
+int ha_mroonga::delete_all_rows()
+{
+  DBUG_ENTER("ha_mroonga::delete_all_rows");
+  DBUG_RETURN((my_errno = HA_ERR_WRONG_COMMAND));
+}
+
+int ha_mroonga::truncate()
+{
+  DBUG_ENTER("ha_mroonga::truncate");
+  DBUG_RETURN(HA_ERR_WRONG_COMMAND);
+}
+
+double ha_mroonga::scan_time()
+{
+  DBUG_ENTER("ha_mroonga::scan_time");
+  DBUG_RETURN(handler::scan_time());
+}
+
+double ha_mroonga::read_time(uint index, uint ranges, ha_rows rows)
+{
+  DBUG_ENTER("ha_mroonga::read_time");
+  DBUG_RETURN(handler::read_time(index, ranges, rows));
+}
+
+const key_map *ha_mroonga::keys_to_use_for_scanning()
+{
+  DBUG_ENTER("ha_mroonga::keys_to_use_for_scanning");
+  DBUG_RETURN(handler::keys_to_use_for_scanning());
+}
+
+ha_rows ha_mroonga::estimate_rows_upper_bound()
+{
+  DBUG_ENTER("ha_mroonga::estimate_rows_upper_bound");
+  DBUG_RETURN(handler::estimate_rows_upper_bound());
+}
+
+void ha_mroonga::update_create_info(HA_CREATE_INFO* create_info)
+{
+  DBUG_ENTER("ha_mroonga::update_create_info");
+  DBUG_RETURN(handler::update_create_info(create_info));
+}
+
+int ha_mroonga::rename_table(const char *from, const char *to)
+{
+  DBUG_ENTER("ha_mroonga::rename_table");
+  DBUG_RETURN(handler::rename_table(from, to));
+}
+
+bool ha_mroonga::is_crashed() const
+{
+  DBUG_ENTER("ha_mroonga::is_crashed");
+  DBUG_RETURN(handler::is_crashed());
+}
+
+bool ha_mroonga::auto_repair() const
+{
+  DBUG_ENTER("ha_mroonga::auto_repair");
+  DBUG_RETURN(handler::auto_repair());
+}
+
+int ha_mroonga::disable_indexes(uint mode)
+{
+  DBUG_ENTER("ha_mroonga::disable_indexes");
+  DBUG_RETURN(HA_ERR_WRONG_COMMAND);
+}
+
+int ha_mroonga::enable_indexes(uint mode)
+{
+  DBUG_ENTER("ha_mroonga::enable_indexes");
+  DBUG_RETURN(HA_ERR_WRONG_COMMAND);
+}
+
+int ha_mroonga::check(THD* thd, HA_CHECK_OPT* check_opt)
+{
+  DBUG_ENTER("ha_mroonga::check");
+  DBUG_RETURN(HA_ADMIN_NOT_IMPLEMENTED);
+}
+
+int ha_mroonga::repair(THD* thd, HA_CHECK_OPT* check_opt)
+{
+  DBUG_ENTER("ha_mroonga::repair");
+  DBUG_RETURN(HA_ADMIN_NOT_IMPLEMENTED);
+}
+
+bool ha_mroonga::check_and_repair(THD *thd)
+{
+  DBUG_ENTER("ha_mroonga::check_and_repair");
+  DBUG_RETURN(TRUE);
+}
+
+int ha_mroonga::analyze(THD* thd, HA_CHECK_OPT* check_opt)
+{
+  DBUG_ENTER("ha_mroonga::analyze");
+  DBUG_RETURN(HA_ADMIN_NOT_IMPLEMENTED);
+}
+
+int ha_mroonga::optimize(THD* thd, HA_CHECK_OPT* check_opt)
+{
+  DBUG_ENTER("ha_mroonga::optimize");
+  DBUG_RETURN(HA_ADMIN_NOT_IMPLEMENTED);
+}
+
+bool ha_mroonga::is_fatal_error(int error_num, uint flags)
+{
+  DBUG_ENTER("ha_mroonga::is_fatal_error");
+  DBUG_RETURN(handler::is_fatal_error(error_num, flags));
+}
+
 #ifdef __cplusplus
 }
 #endif
