@@ -1,6 +1,7 @@
 /* 
   Copyright(C) 2010 Tetsuro IKEDA
   Copyright(C) 2010 Kentoku SHIBA
+  Copyright(C) 2011 Kouhei Sutou <kou@clear-code.com>
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
@@ -100,6 +101,8 @@ class ha_mroonga: public handler
   bool fast_order_limit;
   bool fast_order_limit_with_index;
 
+  bool ignoring_duplicated_key;
+
 public:
   ha_mroonga(handlerton *hton, TABLE_SHARE *share);
   ~ha_mroonga();
@@ -124,6 +127,7 @@ public:
   int rnd_next(uchar *buf);                                        // required
   int rnd_pos(uchar *buf, uchar *pos);                             // required
   void position(const uchar *record);                              // required
+  int extra(enum ha_extra_function operation);
 
   int delete_table(const char *name);
   int write_row(uchar *buf);
@@ -165,6 +169,35 @@ public:
 
   int reset();
 
+  handler *clone(MEM_ROOT *mem_root);
+  uint8 table_cache_type();
+  int read_multi_range_first(KEY_MULTI_RANGE **found_range_p,
+                             KEY_MULTI_RANGE *ranges,
+                             uint range_count,
+                             bool sorted,
+                             HANDLER_BUFFER *buffer);
+  int read_multi_range_next(KEY_MULTI_RANGE **found_range_p);
+  void start_bulk_insert(ha_rows rows);
+  int end_bulk_insert();
+  int delete_all_rows();
+  int truncate();
+  double scan_time();
+  double read_time(uint index, uint ranges, ha_rows rows);
+  const key_map *keys_to_use_for_scanning();
+  ha_rows estimate_rows_upper_bound();
+  void update_create_info(HA_CREATE_INFO* create_info);
+  int rename_table(const char *from, const char *to);
+  bool is_crashed() const;
+  bool auto_repair() const;
+  int disable_indexes(uint mode);
+  int enable_indexes(uint mode);
+  int check(THD* thd, HA_CHECK_OPT* check_opt);
+  int repair(THD* thd, HA_CHECK_OPT* check_opt);
+  bool check_and_repair(THD *thd);
+  int analyze(THD* thd, HA_CHECK_OPT* check_opt);
+  int optimize(THD* thd, HA_CHECK_OPT* check_opt);
+  bool is_fatal_error(int error_num, uint flags);
+
 private:
   void check_count_skip(key_part_map start_key_part_map,
                         key_part_map end_key_part_map, bool fulltext);
